#include <algorithm>
#include <iostream>
#include <vector>
#include <utility>

#include "Util.h"

using namespace std;

template<typename T>
void BubbleSort(vector<T>& v) {
  for (size_t i = v.size() - 1; i > 0; --i) {
    bool swapped = false;
    for (size_t j = 0; j < i; ++j) {
      if (v[j] > v[j + 1]) {
        swap(v[j], v[j + 1]);
        swapped = true;
      }
    }

    if (! swapped) {
      return;
    }
  }
}

template<typename T>
void SelectionSort(vector<T>& v) {
  // Idea: go through the array, find the largest element and put it to the last
  // slot, then rinse & repeat for the rest of the array
  for (size_t i = v.size() - 1; i > 0; --i) {
    size_t largest_index = 0;
    for (size_t j = 1; j <= i; ++j) {
      if (v[j] >= v[largest_index]) {
        largest_index = j;
      }
    }
    if (largest_index != i) {
      swap(v[largest_index], v[i]);
    }
  }
}

// 1 4 3 2 6 5
// 1 3 4 2 ...
// 1 3 _ 4 ... ; 2
// 1 2 3 4 6 5

template<typename T>
void InsertionSort(vector<T>& v) {
  // Idea: sort the prefix of the array, then "insert" the current element to
  // the appropriate slot in the prefix.


  // The correct position we want to find
  for (size_t i = 1; i < v.size(); ++i) {
    ssize_t where_to_put = i - 1;
    // move out the previous value because it will be overwritten
    T item = move(v[i]);
  
    // shift the array while looking for where to put the value
    while (where_to_put >= 0 && v[where_to_put] > item) {
      v[where_to_put + 1] = v[where_to_put];
      where_to_put -= 1;
    }
    // insert the value back
    v[where_to_put + 1] = move(item);
  
    // we performed "i - where_to_put + 1" moves/swaps, record it
    n_swaps_ += i - where_to_put + 1;
  }
}

template<typename T>
void MysterySort(vector<T>& v) {
  for (size_t i = 0; i < v.size(); ++i) {
    for (size_t j = 0; j < v.size(); ++j) {
      if (v[i] < v[j]) {
        swap(v[i], v[j]);
      }
    }
  }
}

template<typename T>
void ImprovedMysterySort(vector<T>& v) {
  for (size_t i = 0; i < v.size(); ++i) {
    for (size_t j = 0; j < i; ++j) {
      if (v[i] < v[j]) {
        swap(v[i], v[j]);
      }
    }
  }
}

// Copy the vector, and run the algorithm & report stats
template<typename T = FancyInt>
void RunAlgorithm(vector<T> v,
                  const std::string & name,
                  function<void(vector<T>&)> sort) {
  Timer timer;
  cout << "Running " << name << "\n";
  ResetStats();
  timer.Tick();
  sort(v);
  timer.Tock();
  ReportStats();
  cout << "elapsed time:   ";
  cout.width(10);
  cout << std::fixed << std::right << timer.ElapsedTime() << " ms\n";
  cout << "---" << endl;
}

int main() {
    vector<int> a{1,2,3,4,5,6,7,8,9,10};
    vector<int> b{1,10,2,9,3,8,4,7,5,6};
    vector<int> c{2,9,4,7,6,5,8,3,10,1};
    vector<int> d{10,9,8,7,6,5,4,3,2,1};
    vector<int> e{1};

    auto sort = InsertionSort<int>;

    PrintVector(a);
    sort(a);
    PrintVector(a);
    cout << "---" << endl;
    PrintVector(b);
    sort(b);
    PrintVector(b);
    cout << "---" << endl;
    PrintVector(c);
    sort(c);
    PrintVector(c);
    cout << "---" << endl;
    PrintVector(d);
    sort(d);
    PrintVector(d);
    cout << "---" << endl;
    PrintVector(e);
    sort(e);
    PrintVector(e);
    cout << "---" << endl;

    const size_t size = 5000;
    
    const auto v = RandomVector<LargeObject>(size);
    // sort a copy so we get the same stats for all algorithms
    // RunAlgorithm<LargeObject>(v, "mystery sort", ImprovedMysterySort<LargeObject>);
    RunAlgorithm<LargeObject>(v, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(v, "selection sort", SelectionSort<LargeObject>);
    RunAlgorithm<LargeObject>(v, "insertion sort", InsertionSort<LargeObject>);

    // generate a sorted list
    vector<LargeObject> sorted_vector;
    for (int i = 0; i < size; ++i) {
      sorted_vector.push_back(i);
    }

    std::cout << "\nsorting an already sorted list\n";
    // RunAlgorithm<LargeObject>(sorted_vector, "mystery sort", ImprovedMysterySort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "selection sort", SelectionSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "insertion sort", InsertionSort<LargeObject>);

    std::reverse(sorted_vector.begin(), sorted_vector.end());
    std::cout << "\nsorting a list sorted in the opposite order\n";
    // RunAlgorithm<LargeObject>(sorted_vector, "mystery sort", ImprovedMysterySort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "selection sort", SelectionSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "insertion sort", InsertionSort<LargeObject>);

    // 1 2 3 4 5

    // 2 3 4 5 1

    return 0;
    
    // an interesting edge case: rotated vector
    std::sort(sorted_vector.begin(), sorted_vector.end());
    std::rotate(sorted_vector.begin(), sorted_vector.begin() + 1, sorted_vector.end());
    std::cout << "\nsorting a list that is almost sorted\n";
    RunAlgorithm<LargeObject>(sorted_vector, "mystery sort", ImprovedMysterySort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "bubble sort", BubbleSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "selection sort", SelectionSort<LargeObject>);
    RunAlgorithm<LargeObject>(sorted_vector, "insertion sort", InsertionSort<LargeObject>);
}
